use anyhow::Context;
use std::path::{Path, PathBuf};
use std::thread::available_parallelism;
use xshell::Shell;

#[derive(Debug, Default, Clone)]
pub(crate) struct SlangBuild {
    source: Option<PathBuf>,
}

impl SlangBuild {
    pub(crate) fn new<P: AsRef<Path>>(source: Option<P>) -> Self {
        Self {
            source: source.map(|path| path.as_ref().to_path_buf()),
        }
    }
}

impl SlangBuild {
    pub(crate) fn assure_source(self: &Self, sh: &Shell) -> anyhow::Result<PathBuf> {
        if self.source.as_ref().is_some_and(|source| source.exists()) {
            return Ok(self.source.as_ref().unwrap().clone());
        }

        let clone_dir = PathBuf::from("target/xtask/slang");
        if clone_dir.exists() {
            return Ok(clone_dir);
        }

        let _ = sh
            .cmd("git")
            .arg("clone")
            .args(["--branch", "v2024.17"])
            .args(["--depth", "1"])
            .arg("--recurse-submodules")
            .arg("--shallow-submodules")
            .arg(format!("-j{}", available_parallelism()?))
            .arg("https://github.com/shader-slang/slang")
            .arg(clone_dir.to_str().unwrap())
            .run()?;
        Ok(clone_dir)
    }

    pub(crate) fn assure_build(self: &Self, sh: &Shell) -> anyhow::Result<PathBuf> {
        let build_dir = sh.create_dir("target/xtask/slang-build")?;
        Ok(build_dir)
    }

    pub(crate) fn configure(self: &Self, sh: &Shell) -> anyhow::Result<PathBuf> {
        let source_dir = self.assure_source(&sh)?;
        let build_dir = self.assure_build(&sh)?;

        sh.cmd("cmake")
            .args(["-S", source_dir.to_str().unwrap()])
            .args(["-B", build_dir.to_str().unwrap()])
            .arg("-DCMAKE_BUILD_TYPE=Release")
            // https://github.com/shader-slang/slang/issues/5832#issuecomment-2533324982
            .arg("-DCMAKE_SKIP_INSTALL_RULES=ON")
            .run()
            .context("slang-build configure")?;

        Ok(PathBuf::from(build_dir))
    }

    pub(crate) fn build_targets(
        self: &Self,
        sh: &Shell,
        targets: &[&str],
    ) -> anyhow::Result<PathBuf> {
        let build_dir = self.configure(&sh)?;
        let cmd = sh
            .cmd("cmake")
            .args(["--build", build_dir.to_str().unwrap()])
            .arg(format!("-j{}", available_parallelism()?));

        targets
            .iter()
            .fold(cmd, |cmd, target| cmd.args(["--target", target]))
            .run()?;

        Ok(build_dir)
    }
}
